@@ -1,30 +1,68 @@
 """
 This module contains a set of common lightweight wrappers to perform simple and non-customised DeltaDeltaCt analyses.
 This is designed for lightweight users that do not wish to employ specialised pipelines.
+The ``qpcr.Pipes`` allow you to "shortcut" some steps for you as they handle object setup for the main ``qpcr`` classes.
+They will perform file-reading, Delta-Ct computation, and normalisation, as well as visualization and saving for you.
 
-> ### A Word on Plotters 
-> Please, note that pipelines fully support mixing "static" and "interactive" Plotters, 
-> but static figures will not stay open if interactive plotters are called to plot after them! 
-> Because `qpcr.Filters` are always called to plot *before* any other `qpcr.Plotters`, this will mainly
-> affect visualising the `qpcr.Plotters.ReplicateBoxplots` generated as Filter-Summaries.
+Which pipeline to use?
+--------------
+
+Depending on which pipeline you choose you will get different levels of customizability. 
+In the basic tutorials we worked with the ``BasicPlus`` pipeline which allows for file-reading, filtering, Delta-Ct computation, and normalisation, as well as visualisation and file-saving for the results.
+It is the "big brother" of the most simple pipeline called ``Basic`` which only performs file-reading, Delta-Ct computation, and normalisation, as well as results saving.
+
+The most "free" pipeline that works with files is the ``Blueprint`` pipeline that allows a user-specified *Reader*, *Analyser*, and *Normaliser*. 
+
+However, these above pipelines were all designed with **multiple regular datafiles** as input in mind! However, as you may have *irregular* or *multi assay files* you may not find these pipelines to be usable for you.
+This is no problem, however, since the ``qpcr.DataReader`` allows for very swift file-reading into ``qpcr.Assay`` objects from many different filetypes. These can then directly be submitted to the ``ddCt`` pipeline. 
+This skips file reading but only performs filtering, Delta-Ct computation, normalisation, and visualisation + file saving. Here is a "pipeline example" using ``ddCt``:
+
+.. code-block:: python
+
+    import qpcr
+    from qpcr.Pipes import ddCt
+
+    myfile = "some_irregular_multi_assay_file.xlsx"
+
+    # read the data externally
+    assays, normalisers = qpcr.read_multi_assay( myfile )
+
+    # now setup the pipeline
+    pipe = ddCt()
+
+    # add our data and run
+    pipe.link( assays = assays, normalisers = normalisers )
+    pipe.run()
+
+    # now get the results to inspect 
+    results = pipe.results()
+
+
+
+Note
+-----
+A Word on Plotters 
+Please, note that pipelines fully support mixing "static" and "interactive" Plotters, 
+but static figures will not stay open if interactive plotters are called to plot after them! 
+Because `qpcr.Filters` are always called to plot *before* any other `qpcr.Plotters`, this will mainly
+affect visualising the `qpcr.Plotters.ReplicateBoxplots` generated as Filter-Summaries.
 """
 
+import logging
 import qpcr
-import matplotlib.pyplot as plt
-import pandas as pd 
-import statistics as stats
 import qpcr._auxiliary.warnings as aw
 import qpcr._auxiliary as aux
 import qpcr.Plotters as Plotters
 import qpcr.Filters as Filters
 import qpcr.Readers as Readers
-import re
-import os 
-import difflib
+import os
+
+logger = aux.default_logger()
+
 
 class Pipeline:
     """
-    This is the basic template class for qpcr Pipelines. 
+    This is the basic template class for qpcr Pipelines.
     It contains a set of basic preliminary methods
     that ensure that elementary required inputs are provided.
 
@@ -32,6 +70,9 @@ class Pipeline:
     ----
     The simplest implementation of this `Pipeline` template is the `Basic` pipeline.
     """
+
+    __slots__ = ['_Normalisers', '_Assays', '_save_to', '_df', '_stats_df', '_Results', '_replicates', '_names', '_softlink']
+
     def __init__(self):
         # super().__init__()
         self._Normalisers = []
@@ -39,7 +80,7 @@ class Pipeline:
         self._save_to = None
         self._df = None
         self._stats_df = None
-        self._Results = None
+        self._Results = qpcr.Results()
         self._replicates = None
         self._names = None
         self._softlink = True
@@ -53,7 +94,7 @@ class Pipeline:
         """
         assays = self._Assays
         return assays
-    
+
     def normalisers(self):
         """
         Returns
@@ -64,35 +105,34 @@ class Pipeline:
         normalisers = self._Normalisers
         return normalisers
 
-    def replicates(self, replicates:(int or tuple)):
+    def replicates(self, replicates: (int or tuple)):
         """
         Set the replicates specifics to use for grouping.
 
         Parameters
         ----------
         replicates : int or tuple
-            Can be an `integer` (equal group sizes, e.g. `3` for triplicates), 
-            or a `tuple` (uneven group sizes, e.g. `(3,2,3)` if the second group is only a duplicate). 
+            Can be an `integer` (equal group sizes, e.g. `3` for triplicates),
+            or a `tuple` (uneven group sizes, e.g. `(3,2,3)` if the second group is only a duplicate).
         """
         self._replicates = replicates
 
-    def names(self, names:(list or dict)):
+    def names(self, names: (list or dict)):
         """
         Set names for replicates groups.
 
         Parameters
         ----------
         names : list or dict
-            Either a `list` (new names without repetitions) or `dict` (key = old name, value = new name) specifying new group names. 
+            Either a `list` (new names without repetitions) or `dict` (key = old name, value = new name) specifying new group names.
             Group names only need to be specified once, and are applied to all replicate entries.
         """
         self._names = names
 
-
     def run(self, **kwargs):
         """
         Will run the pipeline provided that at least minimal inputs have been provided (i.e. Assays and Normalisers, as well as replicate specifics have been provided).
-        This is a wrapper, the actual functional core is defined in the method `self._run()`. 
+        This is a wrapper, the actual functional core is defined in the method `self._run()`.
         To implement your own pipeline, make sure to define your own `_run()`, redefine `run()`
         as you require.
 
@@ -104,11 +144,13 @@ class Pipeline:
 
         # vet if there are at least one normaliser and assay present
         if self._Normalisers == [] or self._Assays == []:
-            aw.HardWarning("Pipeline:no_data")
+            e = aw.PipeError("no_data")
+            logger.critical(e)
+            raise e
 
         self._run(**kwargs)
-    
-    def save_to(self, directory:str):
+
+    def save_to(self, directory: str):
         """
         Set the location where to save result files
 
@@ -126,7 +168,7 @@ class Pipeline:
         """
         Returns
         -------
-        data 
+        data
             A pandas dataframe either in replicate version `kind="df"`
             or in stats version `kind="stats"` (default). Alternatively,
             a qpcr.Results object can be returned using `kind="obj"`.
@@ -135,13 +177,13 @@ class Pipeline:
             return self._stats_df
         elif kind == "df":
             df = self._df
-            if "assay" in df.columns: 
-                df = df.drop(columns = ["assay"])
+            if "assay" in df.columns:
+                df = df.drop(columns=["assay"])
             return df
         elif kind == "obj":
             return self._Results
-    
-    def results( self ):
+
+    def results(self):
         """
         Returns
         -------
@@ -151,11 +193,10 @@ class Pipeline:
         """
         return self._Results
 
-
-    def link(self, assays:(list or str) = None, normalisers:(list or str) = None):
+    def link(self, assays: (list or str) = None, normalisers: (list or str) = None):
         """
-        Links new assays-of-interest / sample assays and/or normaliser assays 
-        to the pipline, either replacing old ones or keeping them, 
+        Links new assays-of-interest / sample assays and/or normaliser assays
+        to the pipline, either replacing old ones or keeping them,
         depending on `softlink()` settings.
 
         Parameters
@@ -170,8 +211,7 @@ class Pipeline:
         self.add_assays(assays)
         self.add_normalisers(normalisers)
 
-
-    def prune(self, assays = True, results = True, normalisers = True):
+    def prune(self, assays=True, results=True, normalisers=True):
         """
         Will clear assays, results, and/or normalisers
 
@@ -179,24 +219,26 @@ class Pipeline:
         ----------
         assays : bool
             Will clear any sample assays in the pipeline if True (default).
-        
+
         results : bool
             Will clear any computed results in the pipline if True (default).
-        
+
         normalisers : bool
             Will clear any normalisers in the pipline if True (default).
         """
-        if assays: self._Assays = []
-        if normalisers: self._Normalisers = []
-        if results: 
+        if assays:
+            self._Assays = []
+        if normalisers:
+            self._Normalisers = []
+        if results:
             self._df = None
             self._stats_df = None
-            self._Results = None
+            self._Results = qpcr.Results()
 
     def add_normalisers(self, normalisers):
         """
         Adds normalisers (filepaths) (keeping any already present)
-        
+
         Parameters
         ----------
         normalisers : list or str
@@ -204,7 +246,7 @@ class Pipeline:
         """
         normalisers = self._from_directory(normalisers)
         self._Normalisers.extend(normalisers)
-    
+
     def add_assays(self, assays):
         """
         Adds assays-of-interest / sample assays (filepaths) (keeping any already present)
@@ -217,10 +259,10 @@ class Pipeline:
         assays = self._from_directory(assays)
         self._Assays.extend(assays)
 
-    def softlink(self, bool = None):
+    def softlink(self, bool=None):
         """
-        If `softlink = True`, then `link_assays()` will 
-        prune any previous assays. Otherwise, it will 
+        If `softlink = True`, then `link_assays()` will
+        prune any previous assays. Otherwise, it will
         add new ones and keep old ones.
 
         Parameters
@@ -243,35 +285,40 @@ class Pipeline:
 
             # if inputs are a directory
             if os.path.isdir(files):
-            
+
                 # get files from the directory
                 datafiles = os.listdir(files)
-            
+
                 # if no files are found, raise error
                 if len(datafiles) == 0:
-                    aw.HardWarning("Pipeline:no_data_input", file = files, traceback = False)
-            
+                    e = aw.PipeError("no_data_input", file=files)
+                    logger.critical(e)
+                    SystemExit(e)
+
                 # combine paths with parent directory
-                datafiles = [os.path.join(files,n) for n in datafiles]
+                datafiles = [os.path.join(files, n) for n in datafiles]
                 return datafiles
 
             # else check if inputs are a single file
             # then just put it into a list to be valid input for .extend()
             elif os.path.isfile(files):
-                
+
                 return [files]
 
-            else: 
-                aw.HardWarning("Pipeline:no_data_input", file = files, traceback = False)
-        
+            else:
+                e = aw.PipeError("no_data_input", file=files)
+                logger.critical(e)
+                SystemExit(e)
+
         # else check if we got a list or tuple of files
         elif isinstance(files, (list or tuple)):
             return files
-        
-        # raise error for anything else...
-        else: 
-            aw.HardWarning("Pipeline:no_data_input", file = files, traceback = False)
 
+        # raise error for anything else...
+        else:
+            e = aw.PipeError("no_data_input", file=files)
+            logger.critical(e)
+            SystemExit(e)
 
     def _run(self, **kwargs):
         """
@@ -282,18 +329,19 @@ class Pipeline:
 
 class Basic(Pipeline):
     """
-    Performs simple standardized DeltaDeltaCt analysis 
+    Performs simple standardized DeltaDeltaCt analysis
     based on two lists of files, one for normaliser assays and one for sample assays.
     This makes use of standard settings for `qpcr.Analyser` and `qpcr.Normaliser`
-    which cannot be customized! 
+    which cannot be customized!
     For customization check out the `Blueprint` pipeline or generate your own.
     """
+
     def __init__(self):
         super().__init__()
-    
+
     def _run(self, **kwargs):
         """
-        The automated standard DeltaDeltaCt pipeline. 
+        The automated standard DeltaDeltaCt pipeline.
         Using default settings of `qpcr.Analyser` and `qpcr.Normaliser`
         """
         reader = qpcr.DataReader()
@@ -301,18 +349,18 @@ class Basic(Pipeline):
         normaliser = qpcr.Normaliser()
 
         # add replicate and names info to kwargs
-        kwargs = dict(kwargs, replicates = self._replicates, names = self._names)
+        kwargs = dict(kwargs, replicates=self._replicates, names=self._names)
 
         # analyse normalisers
-        normalisers = [  reader.read(i, **kwargs) for i in self._Normalisers  ]     
-        normalisers = [  analyser.pipe(i) for i in normalisers  ]
-       
+        normalisers = [reader.read(i, **kwargs) for i in self._Normalisers]
+        normalisers = [analyser.pipe(i) for i in normalisers]
+
         # analyse sample assays
-        assays = [  reader.read(i, **kwargs) for i in self._Assays  ]   
-        assays = [  analyser.pipe(i) for i in assays  ]
+        assays = [reader.read(i, **kwargs) for i in self._Assays]
+        assays = [analyser.pipe(i) for i in assays]
 
         # now normalise
-        normaliser.link( assays = assays, normalisers = normalisers )
+        normaliser.link(assays=assays, normalisers=normalisers)
         normaliser.normalise()
 
         # get and store results
@@ -324,17 +372,19 @@ class Basic(Pipeline):
         if self._save_to is not None:
             results.save(self._save_to)
 
+
 class BasicPlus(Basic):
     """
     The same as the Basic Pipeline, but has the option to integrate Plotters and Filters.
     """
+
     def __init__(self):
         super().__init__()
         self._Plotters = []
         self._Figures = []
         self._Filters = []
-    
-    def prune(self, assays = True, results = True, normalisers = True, figures = True, plotters = False, filters = False):
+
+    def prune(self, assays=True, results=True, normalisers=True, figures=True, plotters=False, filters=False):
         """
         Will clear the pipeline.
 
@@ -342,10 +392,10 @@ class BasicPlus(Basic):
         ----------
         assays : bool
             Will clear any sample assays in the pipeline if True (default).
-        
+
         results : bool
             Will clear any computed results in the pipline if True (default).
-        
+
         normalisers : bool
             Will clear any normalisers in the pipline if True (default).
         figures : bool
@@ -355,29 +405,29 @@ class BasicPlus(Basic):
         filters : bool
             Will clear any Filters in the pipline if True (default False).
         """
-        super().prune(assays = assays, results = results, normalisers = normalisers)
-        if figures: 
+        super().prune(assays=assays, results=results, normalisers=normalisers)
+        if figures:
             self._Figures = []
         if filters:
             self._Filters = []
-        if plotters: 
+        if plotters:
             self._Plotters = []
 
-    def add_plotters(self, *Plotters:object):
+    def add_plotters(self, *Plotters: object):
         """
         Adds already specified qpcr.Plotter objects to the Pipeline.
-        
+
         Parameters
         ----------
         *Plotters
             Any number of qpcr.Plotters.Plotter objects.
         """
         self._Plotters.extend(Plotters)
-    
-    def add_filters(self, *Filters:object):
+
+    def add_filters(self, *Filters: object):
         """
         Adds already specified qpcr.Filter instances to the Pipeline.
-        
+
         Parameters
         ----------
         *Filters
@@ -396,11 +446,11 @@ class BasicPlus(Basic):
 
     def _run(self, **kwargs):
         """
-        The automated standard DeltaDeltaCt pipeline. 
+        The automated standard DeltaDeltaCt pipeline.
         Also produces applies and Plotters to produce figures...
         """
 
-        # simply set up the Blueprint pipeline 
+        # simply set up the Blueprint pipeline
         # and run with default settings
         pipeline = Blueprint()
 
@@ -419,25 +469,24 @@ class BasicPlus(Basic):
         pipeline.run()
 
         # and store the results
-        self._Results = pipeline.get(kind = "obj")
-        self._df = pipeline.get(kind = "df")
-        self._stats_df = pipeline.get(kind = "stats")
-
-
+        self._Results = pipeline.get(kind="obj")
+        self._df = pipeline.get(kind="df")
+        self._stats_df = pipeline.get(kind="stats")
 
 
 class Blueprint(BasicPlus):
     """
-    Performs simple Delta-Delta-Ct analysis based on the same workflow as the `BasicPlus` pipeline, but allows full costumization of SampleReader, Analyser, and Normaliser objects.
-    Optionally, `qpcr.SampleReader`, `qpcr.Analyser`, and `qpcr.Normaliser` may be set up externally and linked into the pipeline. Any non-linked processing classes will be set up using defaults.
+    Performs simple Delta-Delta-Ct analysis based on the same workflow as the `BasicPlus` pipeline, but allows full costumization of Reader, Analyser, and Normaliser objects.
+    Optionally, `Reader`, `qpcr.Analyser`, and `qpcr.Normaliser` may be set up externally and linked into the pipeline. Any non-linked processing classes will be set up using defaults.
     """
+
     def __init__(self):
         super().__init__()
         self._Reader = None
         self._Analyser = None
         self._Normaliser = None
 
-    def prune(self, assays = True, results = True, normalisers = True, figures = True, plotters = False, filters = False, cores = False, reader = False, analyser = False, normaliser = False):
+    def prune(self, assays=True, results=True, normalisers=True, figures=True, plotters=False, filters=False, cores=False, reader=False, analyser=False, normaliser=False):
         """
         Will clear the pipeline.
 
@@ -445,10 +494,10 @@ class Blueprint(BasicPlus):
         ----------
         assays : bool
             Will clear any sample assays in the pipeline if True (default).
-        
+
         results : bool
             Will clear any computed results in the pipline if True (default).
-        
+
         normalisers : bool
             Will clear any normalisers in the pipline if True (default).
         figures : bool
@@ -468,25 +517,14 @@ class Blueprint(BasicPlus):
             Note, that clearing `results` will also clear the `Normaliser`'s results,
             but keep the Normaliser itself!
         """
-        super().prune(
-                        assays = assays, 
-                        results = results, 
-                        normalisers = normalisers,
-                        figures = figures, 
-                        plotters = plotters, 
-                        filters = filters
-                    )
-        
-        try: 
-            self._Normaliser.prune(
-                                    assays = assays, 
-                                    normalisers = normalisers, 
-                                    results = results
-                                )
-        except: 
-            pass
-        
-        if cores: 
+        super().prune(assays=assays, results=results, normalisers=normalisers, figures=figures, plotters=plotters, filters=filters)
+
+        try:
+            self._Normaliser.prune(assays=assays, normalisers=normalisers, results=results)
+        except Exception as e:
+            logger.debug(e)
+
+        if cores:
             self._Reader = None
             self._Analyser = None
             self._Normaliser = None
@@ -498,20 +536,20 @@ class Blueprint(BasicPlus):
             if normaliser:
                 self._Normaliser = None
 
-    def Reader(self, Reader : qpcr.SampleReader = None):
+    def Reader(self, Reader=None):
         """
-        Links a `qpcr.SampleReader` object to the pipeline.
+        Links one of the `qpcr.Readers` to the pipeline.
 
         Parameters
         ----------
-        Reader : qpcr.SampleReader
-            A `qpcr.SampleReader` object
+        Reader
+            A `qpcr.Readers` object
         """
-        if Reader is not None: 
+        if Reader is not None:
             self._Reader = Reader
         return self._Reader
 
-    def Analyser(self, Analyser : qpcr.Analyser = None):
+    def Analyser(self, Analyser: qpcr.Analyser = None):
         """
         Links a `qpcr.Analyser` object to the pipeline.
 
@@ -524,7 +562,7 @@ class Blueprint(BasicPlus):
             self._Analyser = Analyser
         return self._Analyser
 
-    def Normaliser(self, Normaliser : qpcr.Normaliser = None):
+    def Normaliser(self, Normaliser: qpcr.Normaliser = None):
         """
         Links a `qpcr.Normaliser` object to the pipeline.
 
@@ -536,57 +574,57 @@ class Blueprint(BasicPlus):
         if Normaliser is not None:
             self._Normaliser = Normaliser
         return self._Normaliser
-    
+
     def _setup_cores(self):
         """
-        Sets SampleReader, Analyser, and Normaliser to defaults, if no external ones were provided...
+        Sets Reader, Analyser, and Normaliser to defaults, if no external ones were provided...
         """
-        if self.Reader() is None: 
+        if self.Reader() is None:
             self.Reader(qpcr.DataReader())
-        if self.Analyser() is None: 
+        if self.Analyser() is None:
             self.Analyser(qpcr.Analyser())
         if self.Normaliser() is None:
             self.Normaliser(qpcr.Normaliser())
 
     def _run(self, **kwargs):
         """
-        The automated standard DeltaDeltaCt pipeline. 
+        The automated standard DeltaDeltaCt pipeline.
         Also produces applies and Plotters to produce figures...
         """
         # setup Analyser, and Normaliser (if none were provided)
         self._setup_cores()
-        
+
         reader = self.Reader()
         analyser = self.Analyser()
         normaliser = self.Normaliser()
 
         # check if we have plotters or filters
-        have_plotters = len(self._Plotters) != 0 
+        have_plotters = len(self._Plotters) != 0
         have_filters = len(self._Filters) != 0
 
         # add replicate and names info to kwargs
-        kwargs = dict(kwargs, replicates = self._replicates, names = self._names)
+        kwargs = dict(kwargs, replicates=self._replicates, names=self._names)
 
         # analyse normalisers
-        normalisers = [  reader.read(i, **kwargs) for i in self._Normalisers  ]
+        normalisers = [reader.read(i, **kwargs) for i in self._Normalisers]
 
         if have_filters:
             for filter in self._Filters:
-                normalisers = [  filter.pipe(i) for i in normalisers  ]
-            
-        normalisers = [  analyser.pipe(i) for i in normalisers  ]
-       
+                normalisers = [filter.pipe(i) for i in normalisers]
+
+        normalisers = [analyser.pipe(i) for i in normalisers]
+
         # analyse sample assays
-        assays = [  reader.read(i, **kwargs) for i in self._Assays  ]
+        assays = [reader.read(i, **kwargs) for i in self._Assays]
 
         if have_filters:
             for filter in self._Filters:
-                assays = [  filter.pipe(i) for i in assays  ]
-        
-        assays = [  analyser.pipe(i) for i in assays  ]
+                assays = [filter.pipe(i) for i in assays]
+
+        assays = [analyser.pipe(i) for i in assays]
 
         # now normalise
-        normaliser.link( assays = assays, normalisers = normalisers )
+        normaliser.link(assays=assays, normalisers=normalisers)
         normaliser.normalise()
 
         # get and store results
@@ -602,9 +640,9 @@ class Blueprint(BasicPlus):
         if have_filters:
             for filter in self._Filters:
                 if self._save_to is not None or filter.report() is not None:
-                    
+
                     # add report location if none was specified...
-                    if filter.report() is None: 
+                    if filter.report() is None:
                         filter.report(self._save_to)
 
                 # save filter summary fig
@@ -634,7 +672,7 @@ class Blueprint(BasicPlus):
             filename = os.path.join(self._save_to, f"{plotter.id()}_{num}.{suffix}")
             if not os.path.exists(filename):
                 break
-            num+=1
+            num += 1
         return filename
 
 
@@ -643,16 +681,17 @@ class _Qupid_Blueprint(Blueprint):
     This is the implementation of the Plueprint pipeline for the Qupid webapp.
     It redefines the file-reading functions with methods compatible with the UploadedFile objects from streamlit.
     """
+
     def __init__(self):
         super().__init__()
 
     def _setup_cores(self):
         """
-        Sets SampleReader, Analyser, and Normaliser to defaults, if no external ones were provided...
+        Sets Reader, Analyser, and Normaliser to defaults, if no external ones were provided...
         """
-        if self.Reader() is None: 
+        if self.Reader() is None:
             self.Reader(qpcr._Qupid_SampleReader())
-        if self.Analyser() is None: 
+        if self.Analyser() is None:
             self.Analyser(qpcr.Analyser())
         if self.Normaliser() is None:
             self.Normaliser(qpcr.Normaliser())
@@ -669,18 +708,18 @@ class _Qupid_Blueprint(Blueprint):
 
 #     Note
 #     -------
-#     This class relies on `qpcr.Parsers` and decorated assays to get its input data. 
+#     This class relies on `qpcr.Parsers` and decorated assays to get its input data.
 #     If your file does not offer this kind of architecture, choose another pipeline.
 
 #     """
 #     def __init__(self):
 #         super().__init__()
 #         self._src = None
-    
+
 #     def link(self, filename : str, **kwargs):
 #         """
-#         Reads a datafile in csv or excel format containing 
-#         multiple decorated datasets and extracts sample and normalisers assays. 
+#         Reads a datafile in csv or excel format containing
+#         multiple decorated datasets and extracts sample and normalisers assays.
 
 #         Parameters
 #         ----------
@@ -697,15 +736,15 @@ class _Qupid_Blueprint(Blueprint):
 
 #         # read the multi-assay datafile
 #         self._Assays, self._Normalisers = reader.pipe(self._src, **kwargs)
-        
-        
+
+
 #     def _run(self, **kwargs):
 #         """
 #         The main workflow
 #         """
 #         # setup Reader, Analyser, and Normaliser (if none were provided)
 #         self._setup_cores()
-        
+
 #         analyser = self.Analyser()
 #         normaliser = self.Normaliser()
 
@@ -742,7 +781,7 @@ class _Qupid_Blueprint(Blueprint):
 #         for filter in self._Filters:
 #             if self._save_to is not None or filter.report() is not None:
 #                 # add report location if none was specified...
-#                 if filter.report() is None: 
+#                 if filter.report() is None:
 #                     filter.report(self._save_to)
 
 #             figs = filter.plot()
@@ -764,16 +803,16 @@ class _Qupid_Blueprint(Blueprint):
 #         """
 #         # check if a sheet_name was specified in case of multi-sheet files...
 #         use_multi_sheet = "sheet_name" not in kwargs
-#         if self.Reader() is None: 
+#         if self.Reader() is None:
 #             if self._is_multisheet() and use_multi_sheet:
 #                 self.Reader(Readers.MultiSheetReader())
-#             else: 
+#             else:
 #                 self.Reader(Readers.MultiReader())
-#         if self.Analyser() is None: 
+#         if self.Analyser() is None:
 #             self.Analyser(qpcr.Analyser())
 #         if self.Normaliser() is None:
 #             self.Normaliser(qpcr.Normaliser())
-    
+
 #     def _is_multisheet(self):
 #         """
 #         Checks if a provided excel file contains multiple sheets
@@ -791,7 +830,8 @@ class _Qupid_Blueprint(Blueprint):
 #     # add_normalisers is disabled
 #     def add_assays(self):
 #         print("To provide a data input file use link()!\nIf you wish to supply separate files for assays-of-interest and normalisers, checkout another Pipeline as this one only works with a single Mlit-Assay file!")
-    
+
+
 class ddCt(Blueprint):
     """
     Performs only Delta-Delta-Ct and requires `qpcr.Assay` objects as inputs.
@@ -802,51 +842,52 @@ class ddCt(Blueprint):
 
     Note
     ----
-    As the pipeline inherits from the `Blueprint` pipeline it does have a `Reader` method (which won't do anything though!). 
+    As the pipeline inherits from the `Blueprint` pipeline it does have a `Reader` method (which won't do anything though!).
     """
+
     def __init__(self):
         super().__init__()
-        
+
     def _setup_cores(self):
         """
         Sets Analyser, and Normaliser to defaults, if no external ones were provided...
         """
-        if self.Analyser() is None: 
+        if self.Analyser() is None:
             self.Analyser(qpcr.Analyser())
         if self.Normaliser() is None:
             self.Normaliser(qpcr.Normaliser())
-        
+
     def _run(self, **kwargs):
         # setup Analyser, and Normaliser (if none were provided)
         self._setup_cores()
-        
+
         analyser = self.Analyser()
         normaliser = self.Normaliser()
 
         # check if we have plotters or filters
-        have_plotters = len(self._Plotters) != 0 
+        have_plotters = len(self._Plotters) != 0
         have_filters = len(self._Filters) != 0
 
         # analyse normalisers
-        normalisers = [  i for i in self._Normalisers  ]
+        normalisers = [i for i in self._Normalisers]
 
         if have_filters:
             for filter in self._Filters:
-                normalisers = [  filter.pipe(i) for i in normalisers  ]
-            
-        normalisers = [  analyser.pipe(i) for i in normalisers  ]
-       
+                normalisers = [filter.pipe(i) for i in normalisers]
+
+        normalisers = [analyser.pipe(i) for i in normalisers]
+
         # analyse sample assays
-        assays = [  i for i in self._Assays  ]
+        assays = [i for i in self._Assays]
 
         if have_filters:
             for filter in self._Filters:
-                assays = [  filter.pipe(i) for i in assays  ]
-        
-        assays = [  analyser.pipe(i) for i in assays  ]
+                assays = [filter.pipe(i) for i in assays]
+
+        assays = [analyser.pipe(i) for i in assays]
 
         # now normalise
-        normaliser.link( assays = assays, normalisers = normalisers )
+        normaliser.link(assays=assays, normalisers=normalisers)
         normaliser.normalise()
 
         # get and store results
@@ -862,9 +903,9 @@ class ddCt(Blueprint):
         if have_filters:
             for filter in self._Filters:
                 if self._save_to is not None or filter.report() is not None:
-                    
+
                     # add report location if none was specified...
-                    if filter.report() is None: 
+                    if filter.report() is None:
                         filter.report(self._save_to)
 
                 # save filter summary fig
@@ -884,6 +925,7 @@ class ddCt(Blueprint):
                     filename = self._make_figure_filename(plotter)
                     plotter.save(filename)
 
+
 if __name__ == "__main__":
 
     norm_files = ["./Examples/Example Data/28S.csv", "./Examples/Example Data/actin.csv"]
@@ -893,9 +935,9 @@ if __name__ == "__main__":
     sample_folder = "./Example Data 3/samples/"
 
     groupnames = ["wt-", "wt+", "ko-", "ko+"]
-    
+
     analysis = Blueprint()
-    
+
     # grouped_analyser = qpcr.Analyser()
     # grouped_analyser.anchor("first")
     # analysis.Analyser(grouped_analyser)
@@ -906,7 +948,7 @@ if __name__ == "__main__":
 
     analysis.add_assays(sample_folder)
     analysis.add_normalisers(norm_folder)
-    
+
     # print("No Reps specified, all inferred!")
     analysis.replicates(6)
     analysis.names(groupnames)
@@ -917,7 +959,7 @@ if __name__ == "__main__":
     iqr_filter.report("Example Data 4")
     analysis.add_filters(iqr_filter)
 
-    preview = Plotters.PreviewResults(mode = "static")
+    preview = Plotters.PreviewResults(mode="static")
     analysis.add_plotters(preview)
 
     # now that pipeline is ready, we can run!
@@ -925,16 +967,16 @@ if __name__ == "__main__":
 
     # now we can get results!
     results = analysis.get(kind="df")
-    print(results)    
+    print(results)
     # norm_folder = "Example Data 3/normalisers"
     # exp_folder = "Example Data 3/experimental/samples"
     # ctr_folder = "Example Data 3/experimental/samples"
 
     # groupnames = ["wt-", "wt+", "ko-", "ko+"]
-    
+
     # analysis = BasicPlus(index = "Example Data 2/index.csv")
     # analysis.save_to("Example Data ")
-    
+
     # print(analysis._index)
     # # analysis.set_experimental(assays = exp_folder, normalisers = norm_folder)
     # # analysis.set_control(assays = ctr_folder, normalisers = norm_folder)
